/* UADE - Unix Amiga Delitracker Emulator
 * Copyright 2000-2001, Heikki Orsila <heikki.orsila@ee.tut.fi>
 * See http://www.ee.tut.fi/~heikki/
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>
#include <dirent.h>

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "events.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "debug.h"

#include "delimsg.h"
#include "uade.h"
#include "uadeconfig.h"

#include "decrunch/decrunch.h"

static FILE *uade_loadfile(char *fullname);
static void uade_putlong(int addr,int val);
static void uade_siginthandler(int);
static void uade_signalhandler(int);
static void uade_getmessage(void);
static void uade_waitformessage(int msg);
static void uade_setsubsongfunction(int);
static void uade_setntsc(int);
static void uade_setsongend(int songendpossible);
static void uade_sendamigamessage(int);

uade_msgstruct* uade_struct;

char uade_playername[256];       // filename of deliplayer
char uade_modulename[256];       // filename of song
char uade_scorename[256];        // filename of score file
char uade_mapname[256];          // filename of sharemem file

char deli_playername[256];       // playername that deliplayer reported (from emulation)
char deli_modulename[256];       // modulename that deliplayer reported
char deli_formatname[256];	 // formatname that deliplayer reported

int uade_loadmoduleboolean = 0;
int uade_setsubsongboolean = 0;
int uade_subsong;
int uade_execdebugboolean = 0;
int uade_force_by_default = 0;
// file descriptor to send stream to an external program
int uade_soundfd;
// is uade commaned by external application with memorymap ?
int uade_slavemode = 0;
// for cmdline tool streaming to another application
int uade_usingoutpipe = 0;

int uade_use_ntsc = 0;

static int uade_songendpossible = 1;

int uade_reboot;

int uade_messages;       // number of messages in msgtable
static int msgtable[20]; // contains signals sent from master

static struct sigaction sigact;
static struct sigaction sigint; /* for CTRL-C */

static int voltestboolean = 0; /* enabled if value is 0x12345678 */

void uade_option(int argc, char** argv) {
  int i,fd;
  const int sharedmem_length=65536;
  FILE *testfile;

  uade_mapname[0]='\0';

  /* find score: priority: 1. current dir 2. home dir 3. global dir */
  strcpy(uade_scorename,"score");
  testfile = fopen(uade_scorename,"r");
  if(testfile) {
    fclose(testfile); testfile = 0;
  } else {
    strcpy(uade_scorename,getenv("HOME"));    
    strcat(uade_scorename,"/.xmms/uade/score");
    testfile = fopen(uade_scorename,"r");
    if(testfile) {
      fclose(testfile); testfile = 0;
    } else {
      strcpy(uade_scorename,UADECONFIG_SCORENAME);
      testfile = fopen(uade_scorename,"r");
      if(testfile) {
	fclose(testfile); testfile = 0;
      } else {
	uade_scorename[0]=0;
      }
    }
  }


  for (i = 1; i < argc;) {

    // check for playername parameter
    if (strcasecmp (argv[i], "-P") == 0) {
      strncpy(uade_playername,argv[i+1],250);
      fprintf(stderr,"-P parameter: %s\n",uade_playername);
      i+=2;

      // check for modulename parameter
    } else if (strcasecmp (argv[i], "-M") == 0) {
      strncpy(uade_modulename,argv[i+1],250);
      fprintf(stderr,"-M parameter: %s\n",uade_modulename);
      uade_loadmoduleboolean = 1;
      i+=2;

      // check for scorename parameter
    } else if(strcasecmp(argv[i],"-S") == 0) {
      strcpy(uade_scorename,argv[i+1]);
      fprintf(stderr,"-S parameter: %s\n",uade_scorename);
      i+=2;

      // check if this prog is requested for slave
      // that let's external programs use this program as
      // a player
    } else if (strcmp (argv[i], "-slave") == 0) {
      strncpy(uade_mapname,argv[i+1],255);
      //      fprintf(stderr,"-slave parameter: %s\n",uade_mapname);
      uade_slavemode = 1;
      i+=2;

    } else if (strcmp (argv[i], "-outpipe") == 0) {
      uade_soundfd = atoi(argv[i+1]);
      fprintf(stderr,"-outpipe parameter: %d\n",uade_soundfd);
      uade_usingoutpipe = 1;
      i+=2;

    } else if (strcmp(argv[i],"-force") == 0) {
      fprintf(stderr,"-force parameter\n");
      uade_force_by_default = 1;
      i++;

    } else if (strcmp(argv[i],"-execdebug") == 0) {
      fprintf(stderr,"-execdebug parameter\n");
      uade_execdebugboolean = 1;
      i++;

    } else if (strcmp(argv[i],"-voltest") == 0) {
      fprintf(stderr,"-voltest parameter\n");
      voltestboolean = 0x12345678;
      i++;

    } else if (strcmp(argv[i],"-ntsc") == 0) {
      fprintf(stderr,"-ntsc parameter (default is PAL)\n");
      uade_use_ntsc = 1;
      i++;

    } else if (strcmp(argv[i],"-noend") == 0) {
      fprintf(stderr,"-noend parameter (deliplayer can't force song to end\n");
      uade_songendpossible = 0;
      i++;

    } else if (strcmp(argv[i],"-sub") == 0) {
      uade_subsong = atoi(argv[i+1]);
      uade_setsubsongboolean = 1;
      fprintf(stderr,"-sub parameter parsed, setting subsong %d...\n",uade_subsong);
      i+=2;

    } else if (strcmp(argv[i],"-int") == 0) {
      memset(&sigint,0,sizeof(struct sigaction));
      sigint.sa_handler = &uade_siginthandler;
      sigaction(SIGINT,&sigint,0);
      i++;

    } else if (strcmp(argv[i],"-varpid") == 0) {
      /* ADDED BY XIGH */
      int pid, fd;
      char pidstr[10];
      char varname[256];
      strcpy(varname,"/var/run/uade.");
      strcat(varname,getenv("USER"));
      strcat(varname,".pid");
      pid = getpid();
      sprintf(pidstr, "%d", pid);
      fd = open(varname, O_RDWR|O_CREAT|O_TRUNC, 0644);
      if (fd < 0) {
	fprintf(stderr, "could not create %s\n",varname);
      } else {
	write(fd, pidstr, strlen(pidstr));
	close(fd);
      }
      i++;

    } else {
      i++;
    }
  }

  if(uade_slavemode==0) {
    if(uade_scorename[0]==0) {
      fprintf(stderr,"uade/uade.c: couldn't locate score file! Exiting...\n");
      exit(EXIT_FAILURE);
    }
  }


  if(uade_slavemode) {

    fd = open(uade_mapname,O_RDWR);
    if(fd < 0) {
      fprintf(stderr,"uade/uade.c: Can not open sharedmem file! Exiting...\n");
      exit(EXIT_FAILURE);
    }
    uade_struct =
      (uade_msgstruct*) mmap(0,sharedmem_length,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
    if(uade_struct == MAP_FAILED) {
      fprintf(stderr,"uade/uade.c: can not mmap shared mem! Exiting...\n");
      exit(EXIT_FAILURE);
    }

    uade_struct->minsubsong = 0;
    uade_struct->maxsubsong = 0;
    uade_reboot = 1;

    strcpy(uade_scorename,uade_struct->scorename);

    // clear messages in msgtable
    uade_messages = 0;
    for(i=0;i<20;i++) msgtable[i]=0;

    // setup a signal handler
    // use signal() on GNU/Linux, and siget() on Solaris
    //     signal(SIGHUP,&uade_signalhandler);
    // use sigaction() instead
    // setup a signal handler for communication with xmms plugin
    // IMPORTANT: sigcat.sa_flags might use TUNING on some UNIXES!
    memset(&sigact,0,sizeof(struct sigaction));
    sigact.sa_handler = &uade_signalhandler;
    sigaction(SIGHUP,&sigact,0);

    // leave a watchdog that checks if XMMS plugin dies (polls plugin
    // task with a kill(pluginpid,0) signal every 0.5 seconds). If XMMS
    // plugin dies, watchdog kills UAE and itself
    {
      pid_t father,child;
      father = getpid();
      child = fork();
      if(child==0) {
	for(;;) {
	  sleep(2);
	  if(uade_struct->masterpid) {
	    if(kill(uade_struct->masterpid,0)<0) break;
	  }
	}
	// remove the shared memory file that is used by uade and xmms plugin
	(void) remove(uade_mapname);
	kill(father,SIGKILL);
	kill(getpid(),SIGKILL);
      }
    }
  }
}

/* catch CTRL-C and exit succesfully */
void uade_siginthandler(int signum) {
  signum = 0;
  exit(0);
}

void uade_signalhandler(int signum) {
  int msg,msgs;
  char a = 0;
  int i;

  msg = uade_struct->touaemsgtype;
  uade_struct->touaemsgtype = 0;
  //fprintf(stderr,"uade_getmessage: Got msg %d\n",msg);

  switch(msg) {
  case UADE_REBOOT:
    if(uade_reboot==1) return;
    uade_reboot=1;
    uade_clearsoundbuffer();
    //fprintf(stderr,"Received UADE_Reboot!\n");
    break;
  case UADE_SETSUBSONG:
    //fprintf(stderr,"Received UADE_SETSUBSONG (%d)\n",uade_struct->subsong);
    uade_setsubsongfunction(uade_struct->subsong);
    uade_struct->setsubsongboolean = 0;
    return;

  case UADE_PLAYERNAME:
  case UADE_MODULENAME:
  case UADE_SCORENAME:
    break;
  case UADE_NTSC:
    uade_setntsc(uade_struct->ntscbit);
    uade_sendamigamessage(UADE_NTSC);
    //fprintf(stderr,"Received msg telling NTSC = %d\n",uade_struct->ntscbit);
    return;
  case UADE_SONGEND:
    uade_songendpossible = uade_struct->songendpossible;
    uade_setsongend(uade_songendpossible);
    //fprintf(stderr,"Received msg telling songendpossible = %d\n",uade_songendpossible);
    return;
  default:
    fprintf(stderr,"uade_getmessage: Msg was not recognized (%d)\n",msg);
    break;
  }

  msgtable[msg]=1;

  // count messages in msgtable
  msgs=0;
  for(i=0;i<20;i++) msgs+=(msgtable[i]==0)?0:1;
  uade_messages=msgs;
  //fprintf(stderr,"uae: getmessage: uade_messages=%d\n",uade_struct->messages);
}

void uade_waitformessage(int msg) {
  int sl=0;
  //  fprintf(stderr,"uade: uade_waitformessage: %d\n",msg);
  for(;;) {
    if(msgtable[msg]!=0) {
      uade_handlemessage(msg);
      break;
    }
    usleep(1000);
    sl+=1000;
  }
  fprintf(stderr,"uade/uade_waitformessage: slept %d\n",sl/1000);

}


void uade_handlemessage(int msg) {
  int i,msgs;
  //  fprintf(stderr,"uade: handlemessage: msg = %d ",msg);
  //  if(msg==UADE_PLAYERNAME) {
  //    fprintf(stderr,"(playername)\n");
  //  } else {
  //    fprintf(stderr,"\n");
  //  }
  msgtable[msg]=0;
  msgs=0;
  for(i=0;i<20;i++)
    msgs += (msgtable[i]==0) ? 0:1;
  uade_messages=msgs;
}


void uade_prerun(void) {
  /* don't load anything under 0x1000 (exec at $1000) */
  int scoreaddress = 0x1000;
  int userstack = 0x7fc00;
  int superstack = 0x7fff0;
  int playeraddr = 0x8000;
  int relocaddr = 0x40000;
  int modaddr = 0x80000;
  int modnameaddr = 0x400;

  FILE* file;
  uae_u8* memp;
  int bytesread = 0;
  // decrunch variables
  FILE* file_cr;
  char *modfile;
  int status; 

  if(uade_slavemode) {

    /* RACE CONDITION BEGINS */
    /* uade executable is ready to receive commands */
    uade_struct->uade_inited_boolean = 1;
    /* sleep() is interrupted by a SIGHUP from XMMS plugin */
    while(uade_struct->loadnewsongboolean == 0)
      sleep(2);
    /* RACE CONDITION ENDS */

    strcpy(uade_playername,uade_struct->playername);
    strcpy(uade_modulename,uade_struct->modulename);
    strcpy(uade_scorename,uade_struct->scorename);
    uade_loadmoduleboolean = (uade_modulename[0]=='\0') ? 0 : 1;

    uade_struct->loadnewsongboolean = 0;

    //fprintf(stderr,"Received playername, modulename and scorename\n");
    //fprintf(stderr,"uade_prerun: got strings: player: %s mod: %s\n",uade_playername,uade_modulename);
    uade_force_by_default = uade_struct->force_by_default;
    if(uade_struct->setsubsongboolean) {
      uade_struct->setsubsongboolean = 0;
      uade_setsubsongboolean = 1;
      uade_subsong = uade_struct->subsong;
    }
    uade_struct->minsubsong = uade_struct->maxsubsong = 0;
    uade_songendpossible = uade_struct->songendpossible;
  }

  /* set songendpossible to amiga memory*/
  uade_setsongend(uade_songendpossible);

  /* this is for debugging purposes (this call can be removed) */
  uade_putlong(0x190,voltestboolean);

  /* open deliplayer */
  file = fopen(uade_playername,"r");
  if(file) {
    if ((status = decrunch (&file, uade_playername)) < 0) {
      fprintf (stderr, "decrunching error...");
      fclose(file);
      goto setupfailure;
    } else {
      memp=get_real_address(playeraddr);
      bytesread = fread(memp,1,100000000,file);
      fprintf(stderr,"deliplayer '%s' (%d bytes)\n",uade_playername,bytesread);
      fclose(file);
      // set player executable address for relocator
      uade_putlong(0x108,playeraddr);
      // set address to relocate player to ...
      uade_putlong(0x10c,relocaddr);
    }
  } else {
setupfailure:
    if(uade_slavemode) {
      uade_reboot = 0;
      uade_struct->songendboolean = 1;
      return;
    }
    exit(-1);
  }

  /* load module (not necessary) */
  // set module start
  uade_putlong(0x100,modaddr);
  uade_putlong(0x104,0); /* set module size to zero */
  uade_putlong(0x128,0); /* mod name address pointer */
  if(uade_loadmoduleboolean) {
    if((file = fopen(uade_modulename,"r"))) {
      //fprintf (stderr, "testing if file is crunched...\n");
      if ((status = decrunch (&file, uade_modulename)) < 0) {
	fprintf (stderr, "decrunching error...");
	fclose(file);
	goto setupfailure;
      } else {
	/* load file*/
	memp=get_real_address(modaddr);
	bytesread = fread(memp,1,100000000,file);
	fprintf(stderr,"module '%s' (%d bytes)\n",uade_modulename,bytesread);
	fclose(file);
	// set module size
	uade_putlong(0x104,bytesread);
	memp=get_real_address(modnameaddr);
	strncpy(memp,uade_modulename,255);
	uade_putlong(0x128,modnameaddr);
      }
    } else {
      goto setupfailure;
    }
  } else {
    // set module size to zero
    uade_putlong(0x104,0);
  }

  /* load sound core (score) */
  if((file = fopen(uade_scorename,"r"))) {
    memp=get_real_address(scoreaddress);
    bytesread = fread(memp,1,100000000,file);
    fprintf(stderr,"score '%s' (%d bytes)\n",uade_scorename,bytesread);
    fclose(file);
    // set Amiga running from begin of the soundcore
    m68k_setpc(scoreaddress);
    // override bit for sound format checking
    uade_putlong(0x118,uade_force_by_default);
    // exec library debugging feature
    if(uade_execdebugboolean) {
      uade_putlong(0x18C,0x12345678);
    } else {
      uade_putlong(0x18C,0);
    }
    // setsubsong
    uade_putlong(0x11c,uade_setsubsongboolean);
    uade_putlong(0x120,uade_subsong);

    /* set ntscbit correctly */
    if(uade_slavemode!=0) {
      uade_setntsc(uade_struct->ntscbit);
    } else {
      uade_setntsc(uade_use_ntsc);
    }
    // pause bits (don't care!), for debugging purposes only
    uade_putlong(0x180,0);
    uade_putlong(0x184,0);
    // set user and supervisor stack pointers
    uade_putlong(0x110,userstack);
    uade_putlong(0x114,superstack);

    // input msg handler flag to zero
    uade_putlong(0x300,0);

  } else {
    goto setupfailure;
  }

  uade_setsubsongboolean = 0;

  deli_playername[0]='\0';
  deli_modulename[0]='\0';
  if(uade_slavemode) {
    uade_struct->deli_playername[0]='\0';
    uade_struct->deli_modulename[0]='\0';
  }

  // don't reboot after this routine has run unless SIGNAL has been
  // caught...
  if(uade_slavemode) {
    uade_reboot = 0;
    if(msgtable[UADE_REBOOT]) {
      int l;
      for(l=0;l<20;l++) {
	fprintf(stderr,"uade/uade.c: TRIED TO REBOOT 2 TIMES !!!\n");
	fprintf(stderr," If you can make this bug happen again please submit");
	fprintf(stderr," a bug report.\n");
      }
    }
  }
}

static void uade_putlong(int addr,int val) {
  uae_u8 *p;
  p=get_real_address(addr+0);
  *p = (uae_u8) (val>>24 & 0xff);
  p=get_real_address(addr+1);
  *p = (uae_u8) (val>>16 & 0xff);
  p=get_real_address(addr+2);
  *p = (uae_u8) (val>>8 & 0xff);
  p=get_real_address(addr+3);
  *p = (uae_u8) (val & 0xff);
}

int uade_getlong(int adr) {
  uae_u8* ptr;
  int x=0;
  ptr=get_real_address(adr);
  x+=(*ptr)<<24;
  ptr=get_real_address(adr+1);
  x+=(*ptr)<<16;
  ptr=get_real_address(adr+2);
  x+=(*ptr)<<8;
  ptr=get_real_address(adr+3);
  x+=(int) (*ptr);
  return x;
}

void uade_getamigamessage(void) {
  uae_u8* ptr;
  uae_u8* nameptr;
  int x;
  int mins,maxs,curs;
  FILE *file;
  int size;
  int status;

  // GET MESSAGE TYPE FROM VIRTUAL ADDR $200
  x = uade_getlong(0x200);

  switch(x) {
  case UADE_SONGEND:
    fprintf(stderr,"uae/uade.c: song end (deliplayer)\n");
    if(uade_slavemode == 0) {
      exit(EXIT_FAILURE);
    } else {
      uade_struct->songendboolean = 1;
    }
    break;
  case UADE_SUBSINFO:
    mins=uade_getlong(0x204);
    maxs=uade_getlong(0x208);
    curs=uade_getlong(0x20C);
    fprintf(stderr,"uae/uade.c: subsong info: min subsong: %d max subsong: %d cur subsong: %d\n",mins,maxs,curs);
    if(uade_slavemode != 0) {
      uade_struct->minsubsong=mins;
      uade_struct->maxsubsong=maxs;
      uade_struct->subsong=curs;
    }
    break;
  case UADE_PLAYERNAME:
    ptr = get_real_address(0x204);
    strncpy(deli_playername,ptr,25);
    fprintf(stderr,"uae/uade.c: playername: %s\n",deli_playername);
    if(uade_slavemode) {
      strncpy(uade_struct->deli_playername,deli_playername,25);
    }
    break;
  case UADE_MODULENAME:
    ptr = get_real_address(0x204);
    strncpy(deli_modulename,ptr,25);
    fprintf(stderr,"uae/uade.c: modulename: %s\n",deli_modulename);
    if(uade_slavemode) {
      strncpy(uade_struct->deli_modulename,deli_modulename,250);
    }
    break;
  case UADE_FORMATNAME:
    ptr = get_real_address(0x204);
    strncpy(deli_formatname,ptr,255);
    fprintf(stderr,"uae/uade.c: formatname: %s\n",deli_formatname);
    if(uade_slavemode) {
      strncpy(uade_struct->deli_formatname,deli_formatname,255);
    }
    break;
  case UADE_GENERALMSG:
    fprintf(stderr,"uae/uade.c: general message: %s\n",get_real_address(0x204));
    break;
  case UADE_CHECKERROR:
    fprintf(stderr,"uae/uade.c: FAILURE: Module check FAILED\n");
    break;
  case UADE_SCORECRASH:
    fprintf(stderr,"uae/uade.c: FAILURE: Soundcore CRASHED\n");
    if(uade_slavemode) uade_struct->songendboolean = 1;
    break;
  case UADE_SCOREDEAD:
    fprintf(stderr,"uae/uade.c: FAILURE: Soundcore is in DEADEND\n");
    if(uade_slavemode!=0) {
      uade_struct->songendboolean = 1;
    }
    break;
  case UADE_LOADFILE:
    /* load a file named at 0x204 (name pointer) to address pointed by
       0x208 and insert the length to 0x20C */
    nameptr = get_real_address(uade_getlong(0x204));
    if((file = uade_loadfile(nameptr))) {

      if ((status = decrunch (&file, nameptr)) < 0) {
	fprintf (stderr, "uae/uade.c: DTP_LoadFile: Decrunching error\n");
	fclose(file);
	break;
      } else {
	ptr = get_real_address(uade_getlong(0x208));
	size = fread(ptr,1,1000000,file);
	fclose(file);
	uade_putlong(0x20C,size);
	fprintf(stderr,"uae/uade.c: DTP_LoadFile: file loaded: ptr = 0x%x size = 0x%x\n",uade_getlong(0x208),size);
      }
    }
    break;

  default:
    fprintf(stderr,"uae/uade.c: UNKNOWN MESSAGE from Amiga (trap #5): ADDR $200: $%x\n",x);
    break;
  }
}

static FILE *uade_loadfile(char *fullname) {
  FILE *file;
  char *separator;
  char *name;     /* points to file without path */
  char path[256]; /* contains dirpath to the file */
  DIR *dir;
  struct dirent *direntry;
  char loadname[256];

  fprintf(stderr,"uae/uade.c: DTP_LoadFile: trying to load file '%s'\n",fullname);
  if ((file=fopen(fullname,"r"))) return file;

  if((separator=strrchr(fullname,(int) '/'))) {
    name = separator + 1;
    *separator=0;
    path[255]=0;
    strncpy(path,fullname,255);
    strncat(path,"/",255);
    *separator='/';
  } else {
    name = fullname;
    strcpy(path,"./");
  }

  if((dir = opendir(path))==0) {
    fprintf(stderr,"uae/uade.c: DTP_LoadFile: Can't open module directory '%s' for scanning!\n",path);
    return 0;
  }

  // fprintf(stderr,"name = '%s' path = '%s'\n",name,path);
  while((direntry = readdir(dir))) {
    // fprintf(stderr,"dirent: %s\n",direntry->d_name);
    if(strcasecmp(name,direntry->d_name)==0) {
      loadname[255]=0;
      strncpy(loadname,path,255);
      strncat(loadname,direntry->d_name,255);
      if((file = fopen(loadname,"r"))) break;
    }
  }
  closedir(dir);

  if(file) {
    fprintf(stderr,"uae/uade.c: DTP_LoadFile: trying to load file '%s'\n",loadname);
  } else {
    fprintf(stderr,"uae/uade.c: DTP_LoadFile: couldn't load file '%s'\n",fullname);
  }

  return file;
}


static void uade_setsubsongfunction(int subs) {
  uade_putlong(0x120,subs);
  uade_sendamigamessage(UADE_SETSUBSONG);
  uade_clearsoundbuffer();
}

static void uade_setntsc(int usentsc) {
  uade_putlong(0x124,usentsc);
}

static void uade_setsongend(int songendpossible) {
  uade_putlong(0x12C,songendpossible);
}

static void uade_sendamigamessage(int msgtype) {
  uade_putlong(0x300,msgtype);
}

void uade_clearsoundbuffer(void) {
  if(uade_slavemode!=0) {
    memset(uade_struct->soundbuffer,0,sizeof(uade_struct->soundbuffer));
  }
}
